from typing import List
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader


import pickle
from . import sier_net


class EasierNetEstimator:
    """ This is the Ensembling by Averaging Sparse-Input Hierarchical Networks(EASIER-NET)"""
    def __init__(
        self,
        n_estimators: int, #Size of ensemble; number of SIER-nets being ensembled.
        input_filter_layer: bool, #Scales the inputs by parameter β
        n_layers: int, #Number of hidden layers
        n_hidden: int, #Number of hidden nodes in each layer
        full_tree_pen: float, #λ2; controls the number of active layers and hidden nodes
        input_pen: float, #λ1; Controls the input sparsity
        batch_size: int, #size of mini-batches for Adam
        num_classes: int, #Equal to 0 if doing a regression, otherwise number of classes for binary / multi-classification
        weight: list,
        max_iters: int, #Number of epochs to run Adam
        max_prox_iters: int, #Number of epochs to run batch proximal gradient descent
        model_fit_params_file: str =None, #A json file that specifies what the hyperparameters are. If given, this will override the arguments passed in.
    ):

        self.num_classes=num_classes
        self.input_filter_layer=input_filter_layer
        self.n_layers=n_layers
        self.n_hidden=n_hidden
        self.n_estimators=n_estimators
        self.max_iters=max_iters
        self.max_prox_iters=max_prox_iters
        self.full_tree_pen=full_tree_pen
        self.input_pen=input_pen
        self.batch_size=batch_size
        self.model_fit_params_file=model_fit_params_file
        self.score_criterion = (
            nn.CrossEntropyLoss() if self.num_classes >= 2 else nn.MSELoss()
        )
        self.weight = weight

        self.estimators = self._generate_estimators()
        if self.model_fit_params_file is not None:
            self.load_model(model_fit_params_file)


        assert num_classes != 1
        assert input_filter_layer

    def _generate_estimators(self) -> List[sier_net.SierNetEstimator]:
        estimators=[]
        for i in range(self.n_estimators):
            est=sier_net.SierNetEstimator(
                input_filter_layer=self.input_filter_layer,
                n_layers= self.n_layers,
                n_hidden= self.n_hidden,
                full_tree_pen=self.full_tree_pen,
                input_pen=self.input_pen,
                batch_size=self.batch_size,
                num_classes=self.num_classes,
                weight=self.weight,
                max_iters=self.max_iters,
                max_prox_iters=self.max_prox_iters,
            )

            estimators.append(est)

        return estimators


    def write_model(self, filename): #save and write model to file
        #save as .pt file
        meta_state_dict = self.get_params()
        meta_state_dict["state_dicts"] = [
            est.net.state_dict() for est in self.estimators
        ]
        for state_dict in meta_state_dict["state_dicts"]:
            state_dict["n_inputs"] = self.n_inputs

        torch.save(meta_state_dict, filename)

    def load_model(self, model_fit_params_file):
        meta_state_dict = torch.load(model_fit_params_file, map_location='cpu') #to avoid gpu ram surge
        state_dict_list = meta_state_dict["state_dicts"]
        for (sier_net, state_dict) in zip(self.estimators, state_dict_list):
            sier_net._load_state_dict(state_dict)

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
    ):
        assert len(y.shape) == 2
        self.n_inputs=x.shape[1]
        for est in self.estimators:
            est.fit(x, y)

        return self

    def score(self, x: np.ndarray, y: np.ndarray) -> float:
        nparr_to_torch = lambda nparr: (
            torch.Tensor(nparr)
            if self.num_classes == 0
            else torch.from_numpy(nparr.astype(int))[:, 0]
        )
        if self.num_classes == 0:
            y_pred = self.predict(x)
            return -np.mean(np.power(y_pred - y, 2))
        elif self.num_classes == 2:
            pred_log_prob = np.log(self.predict_proba(x))
            log_lik = pred_log_prob[np.arange(y.size),y.flatten()]
            return -np.mean(log_lik)
        else:
            raise NotImplementedError("didn't implement scoring function for multiclass yet")

    def predict(self, x: np.ndarray) -> np.ndarray:
        if self.num_classes == 0: #if this is regression, return avg
            all_outputs = [est.predict(x) for est in self.estimators]
            return np.mean(all_outputs, axis=0)
        else:
            max_class = np.argmax(self.predict_proba(x), axis=1).reshape((-1,1))
            return max_class

    def predict_proba(self, x: np.ndarray) -> np.ndarray:
        assert self.num_classes > 1
        all_outputs = np.array([est.predict_proba(x) for est in self.estimators])
        mean_probs = np.mean(all_outputs, axis=0)
        return mean_probs

    def get_params(self, deep=False) -> dict:
        return{
            "num_classes": self.num_classes,
            "input_filter_layer": self.input_filter_layer,
            "n_layers": self.n_layers,
            "n_hidden": self.n_hidden,
            "n_estimators": self.n_estimators,
            "input_pen": self.input_pen,
            "full_tree_pen": self.full_tree_pen,
            "max_iters": self.max_iters,
            "max_prox_iters": self.max_prox_iters,
            "batch_size": self.batch_size,
            "weight": self.weight,
        }

    def set_params(self, **param_dict):
        if "num_classes" in param_dict:
            self.num_classes = param_dict["num_classes"]
        if "input_filter_layer" in param_dict:
            self.input_filter_layer = param_dict["input_filter_layer"]
        if "n_layers" in param_dict:
            self.n_layers = param_dict["n_layers"]
        if "n_hidden" in param_dict:
            self.n_hidden = param_dict["n_hidden"]
        if "n_estimators" in param_dict:
            self.n_estimators = param_dict["n_estimators"]
        if "input_pen" in param_dict:
            self.input_pen = param_dict["input_pen"]
        if "full_tree_pen" in param_dict:
            self.full_tree_pen = param_dict["full_tree_pen"]
        if "max_iters" in param_dict:
            self.max_iters = param_dict["max_iters"]
        if "max_prox_iters" in param_dict:
            self.max_prox_iters = param_dict["max_prox_iters"]
        if "batch_size" in param_dict:
            self.batch_size = param_dict["batch_size"]
        if "weight" in param_dict:
            self.weight = param_dict["weight"]

        return self



import pandas as pd
import numpy as np
from sklearn.datasets import load_boston


def normalize_cols(mat, num_train):
    for col in range(mat.shape[1]):
        col_train_mean = mat[:num_train, col].mean()
        col_train_sd = np.sqrt(np.var(mat[:num_train, col]))
        mat[:, col] -= col_train_mean
        if col_train_sd > 0:
            mat[:, col] /= col_train_sd


"""
Regression datas
"""
np.random.seed(0)

"""
CT slices
"""
out_train_file = "_output/ct_slices_train.npz"
out_test_file = "_output/ct_slices_test.npz"
df = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/00206/slice_localization_data.zip"
)
print(df)
mat = df.to_numpy()[:, 1:]
np.random.shuffle(mat)
num_train = mat.shape[0] // 3 * 2

normalize_cols(mat, num_train)

# Predict sale price
np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1],
    y=mat[:num_train, -1:],
    true_y=mat[:num_train, -1:],
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1],
    y=mat[num_train:, -1:],
    true_y=mat[num_train:, -1:],
)

"""
crime
"""
out_train_file = "_output/crime_train.npz"
out_test_file = "_output/crime_test.npz"
df = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/communities/communities.data",
    names=range(128),
)
mat = df.to_numpy()[:, 5:]
np.random.shuffle(mat)
num_train = mat.shape[0] // 3 * 2

good_cols = np.where(
    [np.mean(mat[:, col] == "?") < 0.1 for col in range(mat.shape[1])]
)[0]
mat = mat[:, good_cols]
mat[mat == "?"] = 0
mat = mat.astype(float)

print(good_cols)
normalize_cols(mat, num_train)

# Predict sale price
np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1],
    y=mat[:num_train, -1:],
    true_y=mat[:num_train, -1:],
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1],
    y=mat[num_train:, -1:],
    true_y=mat[num_train:, -1:],
)

"""
Boston
"""
out_train_file = "_output/boston_train.npz"
out_test_file = "_output/boston_test.npz"
X, y = load_boston(return_X_y=True)
print(y.shape, X.shape)
mat = np.hstack([X, y.reshape((-1, 1))])
np.random.shuffle(mat)
num_train = mat.shape[0] // 3 * 2

normalize_cols(mat, num_train)

# Predict sale price
np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1],
    y=mat[:num_train, -1:],
    true_y=mat[:num_train, -1:],
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1],
    y=mat[num_train:, -1:],
    true_y=mat[num_train:, -1:],
)

"""
Residential Building Data Set Data Set

sale pice last column,
construction cost second to last column
"""
out_train_file = "_output/iran_house_train.npz"
out_test_file = "_output/iran_house_test.npz"
df = pd.read_excel(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/00437/Residential-Building-Data-Set.xlsx",
    header=1,
)
mat = df.to_numpy()
np.random.shuffle(mat)
num_train = mat.shape[0] // 3 * 2

normalize_cols(mat, num_train)

# Predict sale price
np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-2],
    y=mat[:num_train, -1:],
    true_y=mat[:num_train, -1:],
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-2],
    y=mat[num_train:, -1:],
    true_y=mat[num_train:, -1:],
)

"""
wine
"""
out_train_file = "_output/wine_train.npz"
out_test_file = "_output/wine_test.npz"

df = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/wine-quality/winequality-red.csv",
    delimiter=";",
)
mat = df.to_numpy()
np.random.shuffle(mat)
print("tot data", mat.shape[0])
num_train = mat.shape[0] // 3 * 2

normalize_cols(mat, num_train)

np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1],
    y=mat[:num_train, -1:],
    true_y=mat[:num_train, -1:],
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1],
    y=mat[num_train:, -1:],
    true_y=mat[num_train:, -1:],
)

"""
gene cancer
"""
out_train_file = "_output/gene_cancer_train.npz"
out_test_file = "_output/gene_cancer_test.npz"

df = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/00401/TCGA-PANCAN-HiSeq-801x20531.tar.gz"
)
# There are only 801 samples. The extra rows are all nan
# First columns is meaningless. Let's try to predict the expr of the second gene in this dataset
df = df.iloc[:801, 2:]
mat = df.to_numpy()
mat = np.nan_to_num(mat)
np.random.shuffle(mat)
num_train = mat.shape[0] // 3 * 2

normalize_cols(mat, num_train)

np.savez_compressed(
    out_train_file,
    x=mat[:num_train, 1:],
    y=mat[:num_train, :1],
    true_y=mat[:num_train, :1],
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, 1:],
    y=mat[num_train:, :1],
    true_y=mat[num_train:, :1],
)



import pandas as pd
import numpy as np
from sklearn.datasets import load_boston


def normalize_cols(mat, num_train, max_col=None):
    col_std = []
    max_col = mat.shape[1] if max_col is None else max_col
    for col in range(max_col):
        col_train_mean = mat[:num_train, col].mean()
        col_train_sd = np.sqrt(np.var(mat[:num_train, col]))
        mat[:, col] -= col_train_mean
        if col_train_sd > 0:
            mat[:, col] /= col_train_sd
        col_std.append([col_train_mean, col_train_sd])
    return np.array(col_std)


def normalize_cols_with_params(mat, col_std, max_col):
    for col in range(max_col):
        col_train_mean = col_std[col, 0]
        col_train_sd = col_std[col, 1]
        mat[:, col] -= col_train_mean
        if col_train_sd > 0:
            mat[:, col] /= col_train_sd


"""
Classification datas
"""
np.random.seed(0)

"""
Arrythmia
"""
out_train_file = "_output/arrhythmia_train.npz"
out_test_file = "_output/arrhythmia_test.npz"
df = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/arrhythmia/arrhythmia.data",
    names=np.arange(280),
).drop(axis=1, index=13)
print(df)
mat = df.to_numpy()
print(mat)
print("bad ?", np.mean(mat == "?"))
mat[mat == "?"] = 0
for col in range(mat.shape[1] - 1):
    mat[:, col] = mat[:, col].astype(float)
mat[:, -1] = mat[:, -1].astype(int) - 1

uniq_labels = np.unique(mat[:, -1])
label_dict = {uniq_labels[i]: i for i in range(uniq_labels.size)}
for i in range(mat.shape[0]):
    mat[i, -1] = label_dict[mat[i, -1]]

np.random.shuffle(mat)
num_train = mat.shape[0] // 4 * 3
assert np.unique(mat[:num_train, -1:]).size == uniq_labels.size
print(f"UNIQ LABELS {uniq_labels.size}")

normalize_cols(mat, num_train, max_col=mat.shape[1] - 1)

np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1].astype(float),
    y=mat[:num_train, -1:].astype(int),
    true_y=mat[:num_train, -1:].astype(int),
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1].astype(float),
    y=mat[num_train:, -1:].astype(int),
    true_y=mat[num_train:, -1:].astype(int),
)

"""
gene cancer
"""
out_train_file = "_output/gene_cancer_class_train.npz"
out_test_file = "_output/gene_cancer_class_test.npz"

df_y = pd.read_csv("../data/TCGA-PANCAN-HiSeq-801x20531/labels.csv", nrows=801,)
y = df_y.iloc[:, 1].to_numpy()
classes = np.unique([str(a) for a in df_y.iloc[:, 1]])
class_dict = {klass: i for i, klass in enumerate(classes)}
y = [class_dict[a] for a in y]
print(y)

df_x = pd.read_csv("../data/TCGA-PANCAN-HiSeq-801x20531/data.csv", nrows=801)
x = df_x.iloc[:801, 1:].to_numpy()
x = np.nan_to_num(x)

mat = np.hstack([x, np.array(y).reshape((-1, 1))])

np.random.shuffle(mat)
num_train = mat.shape[0] // 3 * 2

normalize_cols(mat, num_train, max_col=mat.shape[1] - 1)
print(np.unique(mat[:num_train, -1:]).astype(int).size, classes.size)
assert np.unique(mat[:num_train, -1:]).astype(int).size == classes.size

np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1].astype(float),
    y=mat[:num_train, -1:].astype(int),
    true_y=mat[:num_train, -1:].astype(int),
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1].astype(float),
    y=mat[num_train:, -1:].astype(int),
    true_y=mat[num_train:, -1:].astype(int),
)

"""
Soybean (large)
"""
out_train_file = "_output/soybean_train.npz"
out_test_file = "_output/soybean_test.npz"
df = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/soybean/soybean-large.data"
)
print(df)
print(df.shape)
mat = df.to_numpy()
classes = np.unique(mat[:, 0])
class_dict = {klass: i for i, klass in enumerate(classes)}
x = mat[:, 1:]
x[x == "?"] = -1
y = [class_dict[a] for a in mat[:, 0]]
mat = np.hstack([x, np.array(y).reshape((-1, 1))]).astype(float)

np.random.shuffle(mat)
num_train = mat.shape[0] // 4 * 3
normalize_cols(mat, num_train, max_col=mat.shape[1] - 1)
print(np.unique(mat[:num_train, -1:]).astype(int).size, classes.size)
assert np.unique(mat[:num_train, -1:]).astype(int).size == classes.size
print(f"UNIQ LABELS {classes.size}")

np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1].astype(float),
    y=mat[:num_train, -1:].astype(int),
    true_y=mat[:num_train, -1:].astype(int),
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1].astype(float),
    y=mat[num_train:, -1:].astype(int),
    true_y=mat[num_train:, -1:].astype(int),
)

"""
Semieon
"""
out_train_file = "_output/semieon_train.npz"
out_test_file = "_output/semieon_test.npz"
df = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/semeion/semeion.data",
    delimiter=" ",
)
mat = df.to_numpy()
x = mat[:, :256]
y = np.where(mat[:, 256:266])[1]
mat = np.hstack([x, y.reshape((-1, 1))])
classes = np.unique(y)

np.random.shuffle(mat)
num_train = mat.shape[0] // 3 * 2
print(np.unique(mat[:num_train, -1:]).astype(int).size, classes.size)
normalize_cols(mat, num_train, max_col=mat.shape[1] - 1)
assert np.unique(mat[:num_train, -1:]).astype(int).size == classes.size

np.savez_compressed(
    out_train_file,
    x=mat[:num_train, :-1].astype(float),
    y=mat[:num_train, -1:].astype(int),
    true_y=mat[:num_train, -1:].astype(int),
)
np.savez_compressed(
    out_test_file,
    x=mat[num_train:, :-1].astype(float),
    y=mat[num_train:, -1:].astype(int),
    true_y=mat[num_train:, -1:].astype(int),
)

"""
Hill-valley
"""
out_train_file = "_output/hill_valley_train.npz"
out_test_file = "_output/hill_valley_test.npz"
df_train = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/hill-valley/Hill_Valley_with_noise_Testing.data"
)
df_test = pd.read_csv(
    "https://archive.ics.uci.edu/ml/machine-learning-databases/hill-valley/Hill_Valley_with_noise_Training.data"
)
mat_train = df_train.to_numpy()
num_train = mat_train.shape[0]
mat_test = df_test.to_numpy()

print("NUM TRA", num_train)
col_std = normalize_cols(mat_train, num_train, max_col=mat_train.shape[1] - 1)
normalize_cols_with_params(mat_test, col_std, max_col=mat_train.shape[1] - 1)

np.savez_compressed(
    out_train_file,
    x=mat_train[:, :-1].astype(float),
    y=mat_train[:, -1:].astype(int),
    true_y=mat_train[:, -1:].astype(int),
)
np.savez_compressed(
    out_test_file,
    x=mat_test[:, :-1].astype(float),
    y=mat_test[:, -1:].astype(int),
    true_y=mat_test[:, -1:].astype(int),
)



import logging
import copy

import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import TensorDataset
from torch.utils.data import DataLoader

from .network import SierNet

class SierNetEstimator:
    """
    This first a single sparse-input hierarchical network (SIER-net).
    """
    def __init__(
        self,
        input_filter_layer:bool,
        n_layers: int ,
        n_hidden: int,
        full_tree_pen: float,
        input_pen: float,
        batch_size: int,
        num_classes: int,
        weight: list,
        max_iters: int,
        max_prox_iters: int,
        connection_pen: float = 0,
        state_dict: dict = None,
    ):
        self.input_filter_layer = input_filter_layer
        self.n_layers = n_layers
        self.n_hidden = n_hidden
        self.n_out = 1 if num_classes == 0 else num_classes

        self.full_tree_pen = full_tree_pen
        self.connection_pen = connection_pen
        self.input_pen = input_pen

        self.batch_size = batch_size

        self.num_classes = num_classes
        assert num_classes != 1
        self.weight = weight

        self.max_iters = max_iters
        self.max_prox_iters = max_prox_iters
        self.criterion = (
            nn.CrossEntropyLoss(weight=torch.Tensor(weight))
            if self.num_classes >= 2
            else nn.MSELoss()
        )

        self.score_criterion = (
            nn.CrossEntropyLoss() if self.num_classes >= 2 else nn.MSELoss()
        )

        if state_dict is not None:
            self._load_state_dict(state_dict, strict=False)
            print("loaded state dict")

    def _load_state_dict(self, state_dict):
        # TODO: fill this out later:
        self.net = SierNet(
            n_input=state_dict["n_inputs"],
            input_filter_layer=self.input_filter_layer,
            n_layers=self.n_layers,
            n_hidden=self.n_hidden,
            n_out=self.n_out,
        )

        self.net.load_state_dict(state_dict, strict=False)

    def run_prox_gd_step(
        self, trainloader, step_sizes=[1e-1, 1e-2, 1e-3, 1e-4, 1e-5, 1e-6]
    ):
        def _soft_threshold(parameter, soft_thres):
            """
            Do soft thresholding (prox operator for lasso)
            """
            soft_thres_mask = torch.abs(parameter.data) < soft_thres
            soft_thres_pos_mask = parameter.data > soft_thres
            soft_thres_neg_mask = parameter.data < -soft_thres
            parameter.data[soft_thres_mask] = 0
            parameter.data[soft_thres_pos_mask] -= soft_thres
            parameter.data[soft_thres_neg_mask] += soft_thres

        optimizer = optim.SGD(self.net.parameters(), lr=1)

        state_dict = copy.deepcopy(self.net.state_dict())
        for i, data in enumerate(trainloader):
            # get the inputs;  is a list of [inputs, labels]
            inputs, labels = data
            labels = labels if self.num_classes == 0 else labels[:, 0]

            for step_size in step_sizes:
                # zero the parameter gradients
                optimizer.zero_grad()

                outputs = self.net.forward(inputs)
                empirical_loss = self.criterion(outputs, labels)
                loss = (
                    empirical_loss
                    + self.input_pen * self.net.input_norm()
                    + self.full_tree_pen * self.net.weight_matrix_norm()
                    + self.connection_pen * self.net.connection_norm()
                )
                empirical_loss.backward()
                # Update params with respect to smooth loss gradient
                for idx, p in enumerate(self.net.parameters()):
                    p.grad *= step_size
                optimizer.step()

                # Now apply prox step
                # Recall that the lasso penalty applies to the bias as well if doing classification
                soft_thres_full_tree = step_size * self.full_tree_pen
                for layer in self.net.layers:
                    if not self.net.is_regression:
                        _soft_threshold(layer.bias, soft_thres_full_tree)
                    _soft_threshold(layer.weight, soft_thres_full_tree)

                if self.input_pen:
                    soft_thres_input = step_size * self.input_pen
                    _soft_threshold(self.net.input_factors, soft_thres_input)
                    _soft_threshold(
                        self.net.input_to_out_layer.weight, soft_thres_input
                    )
                    if not self.net.is_regression:
                        _soft_threshold(
                            self.net.input_to_out_layer.bias, soft_thres_input
                        )

                soft_thres_connection = step_size * self.connection_pen
                _soft_threshold(self.net.connection_factors, soft_thres_connection)

                # reevaluate
                optimizer.zero_grad()
                outputs = self.net.forward(inputs)
                new_empirical_loss = self.criterion(outputs, labels)
                new_loss = (
                    new_empirical_loss
                    + self.input_pen * self.net.input_norm()
                    + self.full_tree_pen * self.net.weight_matrix_norm()
                    + self.connection_pen * self.net.connection_norm()
                )
                if new_loss < loss:
                    logging.info(
                        f"success w step size {step_size}. new loss {new_loss} old loss {loss}"
                    )
                    return True, new_loss.item(), empirical_loss.item()
                else:
                    logging.info(
                        f"try smaller step size. this step size {step_size} didnt work"
                    )
                    # Reset model parameters
                    self.net.load_state_dict(state_dict)

        return False, loss.item(), empirical_loss.item()

    def run_epoch(self, optimizer, trainloader):
        for i, data in enumerate(trainloader):
            # get the inputs;  is a list of [inputs, labels]
            inputs, labels = data
            labels = labels if self.num_classes == 0 else labels[:, 0]

            # zero the parameter gradients
            optimizer.zero_grad()

            outputs = self.net.forward(inputs)
            empirical_loss = self.criterion(outputs, labels)
            loss = (
                empirical_loss
                + self.input_pen * self.net.input_norm()
                + self.full_tree_pen * self.net.weight_matrix_norm()
                + self.connection_pen * self.net.connection_norm()
            )
            loss.backward()

            optimizer.step()
        return loss.item(), empirical_loss.item()

    def get_pen_loss(self, x, y):
        # Assemble data
        torch_y = (
            torch.Tensor(y)
            if self.num_classes == 0
            else torch.from_numpy(y.astype(int))
        )
        dataset = TensorDataset(torch.Tensor(x), torch_y)

        trainloader = DataLoader(dataset, batch_size=x.shape[0], shuffle=True)
        for i, data in enumerate(trainloader):
            # get the inputs;  is a list of [inputs, labels]
            inputs, labels = data
            labels = labels if self.num_classes == 0 else labels[:, 0]

            outputs = self.net.forward(inputs)
            empirical_loss = self.criterion(outputs, labels)
            loss = (
                empirical_loss
                + self.input_pen * self.net.input_norm()
                + self.full_tree_pen * self.net.weight_matrix_norm()
                + self.connection_pen * self.net.connection_norm()
            )
        return loss.item()

    def get_fisher(self, optimizer, trainloader):
        # param_list = [p for layer in self.layers for p in layer.parameters()] + [self.input_factors] + self.input_to_out_layer.parameters() + [self.connection_factors]
        fisher = 0
        num_obs = len(trainloader)
        for i, data in enumerate(trainloader):
            # get the inputs;  is a list of [inputs, labels]
            inputs, labels = data
            labels = labels if self.num_classes == 0 else labels[:, 0]

            # zero the parameter gradients
            optimizer.zero_grad()

            outputs = self.net.forward(inputs)
            empirical_loss = self.criterion(outputs, labels)
            loss = (
                empirical_loss
                + self.input_pen * self.net.input_norm()
                + self.full_tree_pen * self.net.weight_matrix_norm()
                + self.connection_pen * self.net.connection_norm()
            )
            loss.backward()
            gradient = np.concatenate(
                [p.grad.detach().numpy().flatten() for p in self.net.parameters()]
            ).reshape((-1, 1))
            fisher += np.matmul(gradient, gradient.T)

        # input_eye = np.eye(self.n_inputs)
        # log_determinant = np.linalg.slogdet(fisher/num_obs + self.input_pen * input_eye + self.full_tree_pen * full_tree_eye)
        log_determinant = np.linalg.slogdet(
            fisher / num_obs + self.input_pen * np.eye(fisher.shape[0])
        )
        logging.info(f"LOG DET {log_determinant}")

    def fit(
        self,
        x: np.ndarray,
        y: np.ndarray,
    ):
        self.n_inputs = x.shape[1]
        self.net = SierNet(
            n_input=self.n_inputs,
            input_filter_layer=self.input_filter_layer,
            n_layers=self.n_layers,
            n_hidden=self.n_hidden,
            n_out=self.n_out,
        )

        torch_y = (
            torch.Tensor(y)
            if self.num_classes == 0
            else torch.from_numpy(y.astype(int))
        )
        dataset = TensorDataset(torch.Tensor(x), torch_y)

        trainloader = DataLoader(dataset, batch_size=self.batch_size, shuffle=True)

        optimizer = optim.Adam(self.net.parameters(), lr=0.001)

        for epoch in range(self.max_iters):  # loop over the set multiple times
            loss, empirical_loss = self.run_epoch(optimizer, trainloader)

            # print statistics
            if epoch % 100 == 0:
                logging.info(f"[{epoch}] loss: {loss}")
                logging.info(f"[{epoch}] empirical: {empirical_loss}")
                print(f"[{epoch}] loss: {loss} empirical: {empirical_loss}")

                self.net.get_net_struct()

        trainloader = DataLoader(dataset, batch_size=x.shape[0])
        for i in range(self.max_prox_iters):
            did_step, loss, empirical_loss = self.run_prox_gd_step(trainloader)
            if i % 100 == 0:
                logging.info(f"[prox {i}] loss: {loss}")
                logging.info(f"[prox {i}] empirical_loss: {empirical_loss}")
                print(f"[prox {i}] loss: {loss} empirical: {empirical_loss}")
                self.net.get_net_struct()

            if not did_step:
                break

        return self #added for .fit

    def score(self, x: np.ndarray, y: np.ndarray) -> float:
        torch_y = (
            torch.Tensor(y)
            if self.num_classes == 0
            else torch.from_numpy(y.astype(int))[:, 0]
        )
        return -self.score_criterion(self.net(torch.Tensor(x)), torch_y).item()

    def predict(self, x: np.ndarray) -> np.ndarray:
        output = self.net(torch.Tensor(x)).detach().numpy()
        if self.num_classes == 0:
            return output
        else:
            return np.argmax(self.predict_proba(x), axis = 0).reshape((-1,1))

    def predict_proba(self, x: np.ndarray) -> np.ndarray:
        assert self.num_classes > 1
        outputs = self.net(torch.Tensor(x)).detach().numpy()
        pred_probs = np.exp(outputs)/np.sum(np.exp(outputs), axis=1, keepdims=True)
        return pred_probs

    def get_params(self, deep=False) -> dict:
        return {
            "n_layers": self.n_layers,
            "input_filter_layer": self.input_filter_layer,
            "n_inputs": self.n_inputs,
            "n_hidden": self.n_hidden,
            "full_tree_pen": self.full_tree_pen,
            "input_pen": self.input_pen,
            "connection_pen": self.connection_pen,
            "batch_size": self.batch_size,
            "num_classes": self.num_classes,
            "max_iters": self.max_iters,
            "max_prox_iters": self.max_prox_iters,
            "weight": self.weight,
        }

    def set_params(self, **param_dict):
        print("SET PARAMS", param_dict)
        if "n_layers" in param_dict:
            self.n_layers = param_dict["n_layers"]
        if "input_filter_layer" in param_dict:
            self.input_filter_layer = param_dict["input_filter_layer"]
        if "n_hidden" in param_dict:
            self.n_hidden = param_dict["n_hidden"]
        if "connection_pen" in param_dict:
            self.connection_pen = param_dict["connection_pen"]
        if "full_tree_pen" in param_dict:
            self.full_tree_pen = param_dict["full_tree_pen"]
        if "input_pen" in param_dict:
            self.input_pen = param_dict["input_pen"]
        if "batch_size" in param_dict:
            self.batch_size = param_dict["batch_size"]
        if "num_classes" in param_dict:
            self.num_classes = param_dict["num_classes"]
        if "max_iters" in param_dict:
            self.max_iters = param_dict["max_iters"]
        if "max_prox_iters" in param_dict:
            self.max_prox_iters = param_dict["max_prox_iters"]
        if "weight" in param_dict:
            self.weight = param_dict["weight"]

        return self



"""
Execute commands locally or via sbatch.
"""

import click
import os
import subprocess
import time

sbatch_prelude = """#!/bin/bash
#SBATCH -o %s/job.out
#SBATCH -e %s/job.err
#SBATCH -c %d
#SBATCH -N 1
#SBATCH --mail-type=ALL
#SBATCH --mail-user=jeanfeng@uw.edu

cd /home/jfeng2/spinn2/code
ml Python
source ../venv/bin/activate
"""


@click.command()
@click.option(
    "--clusters",
    default="local",
    help="Clusters to submit to. Default is local execution.",
)
@click.argument("target")
@click.argument("to_execute_str")
def cli(clusters, target, to_execute_str):
    """
    Execute a command with targets, perhaps on a SLURM
    cluster via sbatch. Wait until the command has completed.

    TARGETS: Output files as a space-separated list.

    TO_EXECUTE_F_STRING: The command to execute
    """
    if clusters == "local":
        # Local execution.
        click.echo("Executing locally:")
        click.echo(to_execute_str)
        return subprocess.check_output(to_execute_str, shell=True)

    n_cores = 4

    # Put the batch script in the directory of the first target.
    execution_dir = os.path.dirname(target)
    script_name = "job.sh"
    script_full_path = os.path.join(execution_dir, script_name)
    sentinel_path = os.path.join(execution_dir, "sentinel.txt")
    with open(script_full_path, "w") as fp:
        fp.write(sbatch_prelude % (execution_dir, execution_dir, n_cores))
        fp.write(to_execute_str + "\n")
        fp.write("touch %s\n" % sentinel_path)

    # Clean up old job log files if they exist
    scratch_job_files = [
        os.path.join(execution_dir, "job.err"),
        os.path.join(execution_dir, "job.out"),
    ]
    for scratch_file in scratch_job_files:
        if os.path.exists(scratch_file):
            os.remove(scratch_file)

    out = subprocess.check_output(
        "sbatch --clusters %s %s" % (clusters, script_full_path), shell=True
    )
    click.echo(out.decode("UTF-8"))

    # Wait until the sentinel file appears, then clean up.
    while not os.path.exists(sentinel_path):
        time.sleep(5)
    os.remove(sentinel_path)


if __name__ == "__main__":
    cli()



import logging
from collections import OrderedDict

import numpy as np
from scipy.stats import bernoulli
from scipy.special import logsumexp

import torch
import torch.optim as optim
import torch.nn as nn
import torch.nn.functional as F


class Net(nn.Module):
    """
    PyTorch implementation of a simple dense neural network
    """
    def __init__(
        self,
        n_layers,
        n_input,
        n_hidden=5,
        n_out=1,
        dropout=0,
        make_output=True,
        input_filter_layer=False,
    ):
        """
        @param n_layers: number of hidden layers
        @param n_input: number of input nodes
        @param n_hidden: number of hidden nodes per layer
        @param n_out: number of output nodes
        @param dropout: dropout rate
        @param make_output: whether or not to make the output layer
        @param input_filter_layer: whether to add an input filter layer
        """
        super().__init__()
        self.input_filter_layer = input_filter_layer
        if input_filter_layer:
            self.input_factors = nn.Parameter(0.5 * torch.ones(n_input))
        else:
            self.input_factors = torch.ones(n_input)

        self.n_input = n_input
        self.n_hidden = n_hidden
        self.n_layers = n_layers
        self.n_out = n_out
        self.middle_layers = OrderedDict()
        for i in range(n_layers):
            self.middle_layers[str(i)] = nn.Linear(n_hidden, n_hidden)
        self.middle_layer_seq = nn.Sequential(self.middle_layers)

        self.input_layer = nn.Linear(n_input, n_hidden)
        self.layers = [self.input_layer] + list(self.middle_layers.values())
        if make_output:
            self.output_layer = nn.Linear(n_hidden, n_out)
            self.layers += [self.output_layer]

        self.dropout = dropout
        if dropout > 0:
            self.dropout_layer = nn.Dropout(p=dropout)

    @property
    def is_regression(self):
        return self.n_out == 1

    def forward(self, x):
        if self.input_filter_layer:
            x = self.input_factors * x
        x = self.input_layer(x)
        for idx, mid_layer in self.middle_layers.items():
            x = mid_layer(F.relu(x))
            if self.dropout > 0 and self.dropout_layer.p > 0:
                x = self.dropout_layer(x)
        x = self.output_layer(x)
        return x

    def predict(self, x: np.ndarray) -> np.ndarray:
        return self.forward(torch.Tensor(x)).detach().numpy()

    def predict_log_proba(self, x: np.ndarray) -> np.ndarray:
        raw_log_probs = F.log_softmax(self.forward(torch.Tensor(x)), 1).detach().numpy()
        return raw_log_probs

    def norm(self):
        return sum([torch.sum(torch.abs(p)) for p in self.parameters()])

    def input_factor_norm(self):
        return torch.norm(self.input_factors, p=1) if self.input_filter_layer else 0

    def input_layer_norm(self):
        return torch.sum(torch.abs(self.input_layer.weight), 0)

    def layer_norms(self):
        return (
            [torch.sum(torch.abs(self.input_layer.weight), 0).detach().numpy()]
            + [
                torch.sum(torch.abs(layer.weight), 0).detach().numpy()
                for layer in self.middle_layers.values()
            ]
            + [torch.sum(torch.abs(self.output_layer.weight), 0).detach().numpy()]
        )

    def weight_matrix_norm(self):
        bias_norm = (
            sum([torch.norm(layer.bias, p=1) for layer in self.middle_layers.values()])
            + torch.norm(self.input_layer.bias, p=1)
            + torch.norm(self.output_layer.bias, p=1)
        )
        weight_norm = (
            sum(
                [torch.norm(layer.weight, p=1) for layer in self.middle_layers.values()]
            )
            + torch.norm(self.input_layer.weight, p=1)
            + torch.norm(self.output_layer.weight, p=1)
        )

        if self.n_out >= 2:
            return bias_norm + weight_norm
        else:
            return weight_norm

    def support(self, threshold: float = 1e-10):
        input_norms = self.input_layer_norm()
        return input_norms > threshold

    def get_net_struct(self, thres: float = 1e-10) -> dict:
        in_factors = self.input_factors.detach().numpy()
        logging.info(f"in factors {in_factors}")
        if self.input_filter_layer:
            input_factors = np.abs(self.input_factors.detach().numpy())
            support_size = np.sum(input_factors > thres)
            logging.info(f"input factors sort {np.flip(np.sort(input_factors))}")
            logging.info(f"input factors argsort {np.flip(np.argsort(input_factors))}")
            logging.info(f"input factors num big {np.sum(input_factors > thres)}")
            logging.info(
                f"input factors where big {np.where(input_factors > thres)[0]}"
            )
        hidden_sizes = []
        all_layer_norms = self.layer_norms()
        for layer_idx, layer_norms in enumerate(all_layer_norms):
            argsorted_norms = np.flip(np.argsort(layer_norms))
            sorted_norms = np.flip(np.sort(layer_norms))
            if layer_idx == 0:
                inner_support_size = np.sum(layer_norms > thres)
                logging.info(f"INPUT layer:norms {layer_norms}")
                logging.info(f"INPUT layer:arg sort layer norms {argsorted_norms}")
                logging.info(f"INPUT layer:sorted layer norms {sorted_norms}")
                logging.info(f"INPUT layer:support size {inner_support_size}")
                if not self.input_filter_layer:
                    support_size = inner_support_size
            else:
                num_hidden = np.sum(layer_norms > thres)
                hidden_sizes.append(num_hidden)
                logging.info(f"layer {layer_idx}:norms {layer_norms}")
                logging.info(
                    f"layer {layer_idx}:arg sort layer norms {argsorted_norms}"
                )
                logging.info(f"layer {layer_idx}:sorted layer norms {sorted_norms}")
                logging.info(f"layer {layer_idx}:num hidden {num_hidden}")
        return {
            "max_layer": self.n_layers,
            "support_size": support_size,
            "hidden_size_avg": np.mean(hidden_sizes),
        }


class SierNet(Net):
    """
    PyTorch implementation of the actual model in SIER-net
    """
    def __init__(
        self,
        n_layers: int,
        n_input: int,
        n_hidden: int = 5,
        n_out: int = 1,
        input_filter_layer: bool = True,
    ):
        """
        @param n_layers: number of hidden layers (at initialization time)
        @param n_input: number of input nodes (at initialization time)
        @param n_hidden: number of hidden nodes per layer (at initialization time)
        @param n_out: number of output nodes
        @param input_filter_layer: whether to add an input filter layer
        """
        super().__init__(
            n_layers,
            n_input,
            n_hidden=n_hidden,
            n_out=n_out,
            make_output=False,
            dropout=0,
            input_filter_layer=input_filter_layer,
        )

        self.middle_layers = OrderedDict()
        for i in range(n_layers):
            self.middle_layers[str(i)] = nn.Linear(n_hidden, n_hidden)
        self.middle_layer_seq = nn.Sequential(self.middle_layers)

        self.input_layer = nn.Linear(n_input, n_hidden)

        self.input_to_out_layer = nn.Linear(n_input, n_out)
        self.to_out_layers = OrderedDict()
        for i in range(n_layers):
            self.to_out_layers[str(i)] = nn.Linear(n_hidden, n_out)
        self.to_out_layers_seq = nn.Sequential(self.to_out_layers)

        self.layers = (
            list(self.middle_layers.values())
            + list(self.to_out_layers.values())
            + [self.input_layer]
        )
        self.connection_factors = nn.Parameter(torch.ones(self.n_layers + 1))

    def _forward(self, x):
        x = self.input_factors * x
        intermed_out = [self.input_layer(x)] + [None for _ in range(self.n_layers)]
        for idx, mid_layer in enumerate(self.middle_layers.values()):
            intermed_out[idx + 1] = mid_layer(F.relu(intermed_out[idx]))

        # skip connection layers
        abs_connection_factors = torch.abs(self.connection_factors)
        tot_connection_factor = torch.sum(abs_connection_factors)
        normalized_connection_factors = abs_connection_factors / tot_connection_factor
        layer_outputs = [
            normalized_connection_factors[0] * self.input_to_out_layer(x)
        ] + [
            normalized_connection_factors[idx + 1] * to_out_layer(intermed_out[idx + 1])
            for idx, to_out_layer in enumerate(self.to_out_layers.values())
        ]

        return layer_outputs

    def forward(self, x):
        """
        @return model prediction
        """
        layer_outputs = self._forward(x)
        return sum(layer_outputs)

    def connection_norm(self):
        """
        @return norm of the skip-connection weights
        """
        return torch.norm(self.connection_factors, p=1)

    def input_norm(self):
        """
        @return the lasso penalty on the weights connected to the inputs
        """
        weight_norm = torch.norm(self.input_factors, p=1) + torch.norm(
            self.input_to_out_layer.weight, p=1
        )
        bias_norm = torch.norm(self.input_to_out_layer.bias, p=1)
        if self.is_regression:
            return weight_norm
        else:
            return bias_norm + weight_norm

    def weight_matrix_norm(self):
        """
        @return the lasso penalty on the network weights that are not connected to the inputs
        """
        bias_norm = sum([torch.norm(layer.bias, p=1) for layer in self.layers])
        weight_norm = sum([torch.norm(layer.weight, p=1) for layer in self.layers])
        if self.is_regression:
            return weight_norm
        else:
            return bias_norm + weight_norm

    def layer_norms(self):
        """
        @return the norm of each layer
        """
        return [torch.sum(torch.abs(self.input_layer.weight), 0).detach().numpy()] + [
            torch.sum(torch.abs(layer.weight), 0).detach().numpy()
            for layer in self.middle_layers.values()
        ]

    def to_out_layer_norms(self):
        """
        @return the norm of weights connected to the output, partitioned by the layer
        """
        return [
            torch.sum(torch.abs(self.input_to_out_layer.weight), 0).detach().numpy()
        ] + [
            torch.sum(torch.abs(layer.weight), 0).detach().numpy()
            for layer in self.to_out_layers.values()
        ]

    def support(self, threshold: float = 1e-10):
        """
        @return the support of the model as determined by the input filter layer
        """
        return np.array(
            torch.abs(self.input_factors).detach().numpy() > threshold, dtype=int
        )

    def get_importance(self, x: np.ndarray) -> list:
        """
        @return the proportion of variance contributed by each layer
        """
        layer_outputs = self._forward(torch.Tensor(x))
        final_output_var = np.var(sum(layer_outputs).detach().numpy(), axis=0)
        layer_importances_raw = [
            np.var(layer.detach().numpy(), axis=0) for layer in layer_outputs
        ]
        layer_importances = [
            np.mean(l_import / final_output_var) for l_import in layer_importances_raw
        ]
        return layer_importances

    def get_net_struct(self, thres: float = 0) -> dict:
        """
        @return a summary of the network structure
        """
        connection_fac = np.abs(self.connection_factors.detach().numpy())
        connection_fac /= np.sum(connection_fac)
        logging.info(f"connection factors {connection_fac}")
        input_factors = np.abs(self.input_factors.detach().numpy())
        support_size = np.sum(input_factors > thres)
        support = input_factors > thres
        logging.info(f"input factors sort {np.flip(np.sort(input_factors))}")
        logging.info(f"input factors argsort {np.flip(np.argsort(input_factors))}")
        logging.info(f"input factors num big {np.sum(input_factors > thres)}")
        logging.info(f"input factors where big {np.where(input_factors > thres)[0]}")
        hidden_sizes = []
        all_layer_norms = self.layer_norms()
        all_to_out_layer_norms = self.to_out_layer_norms()
        for layer_idx, layer_norms in enumerate(all_layer_norms):
            # if np.abs(connection_fac[layer_idx]) < thres:
            #    continue

            argsorted_norms = np.flip(np.argsort(layer_norms))
            sorted_norms = np.flip(np.sort(layer_norms))
            my_to_out_layer_norms = (
                all_to_out_layer_norms[layer_idx]
                if layer_idx < len(all_to_out_layer_norms)
                else None
            )
            to_out_layer_norms = (
                np.flip(np.sort(all_to_out_layer_norms[layer_idx]))
                if layer_idx < len(all_to_out_layer_norms)
                else None
            )
            argsorted_to_out_norms = (
                np.flip(np.argsort(all_to_out_layer_norms[layer_idx]))
                if layer_idx < len(all_to_out_layer_norms)
                else None
            )

            if layer_idx == 0:
                inner_support_size = np.sum(layer_norms > thres)
                logging.info(f"INPUT layer:norms {layer_norms}")
                logging.info(f"INPUT layer:arg sort layer norms {argsorted_norms}")
                logging.info(f"INPUT layer:sorted layer norms {sorted_norms}")
                logging.info(f"INPUT layer:support size {inner_support_size}")
                to_out_support_size = np.sum(to_out_layer_norms > thres)
                logging.info(f"INPUT to out layer:norms {to_out_layer_norms}")
                logging.info(
                    f"INPUT to out layer:arg sort layer norms {argsorted_to_out_norms}"
                )
                logging.info(f"INPUT to out layer:support size {to_out_support_size}")
                logging.info(
                    f"INPUT to out layer:support {np.where(my_to_out_layer_norms > thres)[0]}"
                )
            else:
                if layer_idx < len(all_to_out_layer_norms):
                    layer_norms += all_to_out_layer_norms[layer_idx]
                num_hidden = np.sum(layer_norms > thres)
                hidden_sizes.append(num_hidden)
                logging.info(f"layer {layer_idx}:sorted layer norms {sorted_norms}")
                logging.info(f"layer {layer_idx}:num hidden {num_hidden}")

        hidden_sizes = np.array(hidden_sizes)
        max_layer = max(
            0,
            min(
                np.max(np.where(connection_fac > thres)[0]),
                np.max(np.where(hidden_sizes > 0)[0]) + 1
                if np.sum(hidden_sizes) > 0
                else 0,
            ),
        )
        connection_fac[max_layer + 1 :] = 0
        connection_fac /= np.sum(connection_fac)
        res = {
            "max_layer": max_layer,
            "support_size": support_size,
            "hidden_size_avg": np.mean(hidden_sizes[hidden_sizes > 0])
            if np.sum(hidden_sizes > 0)
            else 0,
        }
        for i in range(connection_fac.size):
            res[f"connect_{i}"] = connection_fac[i]
        for i in range(hidden_sizes.size):
            res[f"hidden_count_{i}"] = hidden_sizes[i]
        return res
